﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Celeste;
using Celeste.Mod;
using Monocle;
using MonoMod.Utils;
using Mono.Cecil.Cil;
using MonoMod.Cil;
using Microsoft.Xna.Framework;


namespace BetterIceWalls
{
    public class BetterIceWalls : EverestModule
    {
        public static BetterIceWalls Instance;
        public BetterIceWalls() { Instance = this; }

        public override Type SettingsType => typeof(BetterIceWallsSettings);
        public static BetterIceWallsSettings Settings => (BetterIceWallsSettings)Instance._Settings;

        public override void Load()
        {
            Everest.Events.Level.OnLoadLevel += Level_OnLoadLevel;
        }

        private void Level_OnLoadLevel(Level level, Player.IntroTypes playerIntro, bool isFromLoader)
        {
            level.Add(new BetterIceWallController());
        }
        public override void Unload()
        {
            Everest.Events.Level.OnLoadLevel -= Level_OnLoadLevel;
        }
    }

    public enum IceWallTypes { Disabled = 0, Shifted = 1, Scaled = 2, Retexture = 3, Optimized = 4 }

    public class BetterIceWallsSettings : EverestModuleSettings
    {
        [SettingName("BetterIceWalls_Toggle")]
        public IceWallTypes Toggle { get; set; } = IceWallTypes.Scaled;

    }

    
    public class BetterIceWallController : Entity
    {

        public BetterIceWallController() : base()
        {
            Tag = Tags.Global; 
            Depth = 0;
        }

        public override void Awake(Scene scene)
        {
            base.Awake(scene);
            foreach (Entity e in Scene.Entities)
            {
                bool f = e is WallBooster;
                DynamicData d = f ? new DynamicData(typeof(WallBooster), e) : new DynamicData(e.GetType(), e);
                if(d.TryGet("tiles", out object v) && d.TryGet("climbBlocker", out object c))
                {
                    if (d.TryGet<Facings>("Facing", out Facings facing))
                    {
                        if (v.GetType() == typeof(List<Sprite>) && c.GetType() == typeof(ClimbBlocker))
                        {
                            e.Add(new BetterIceWallComponent(f ? typeof(WallBooster) : e.GetType())); //Clever coding makes this most likely be added to this object after all of the sprites are built. therefore we can do some more magic in the component
                        }
                    }
                }
            }
        }
    }
    
    public class BetterIceWallComponent : Component
    {
        private DynamicData dyn;
        private Type type;
        private int oldDepth;
        private Dictionary<int, Dictionary<string, Sprite.Animation>> modifiedSprites = new Dictionary<int, Dictionary<string, Sprite.Animation>>();

        private static Sprite[] mainSprites = new Sprite[3];
        private const float Scale = 0.75f;

        public BetterIceWallComponent(Type t) : base(true, true) {
            type = t;
            if (mainSprites[0] == null)
            {
                mainSprites[0] = GFX.SpriteBank.Create("WallBoosterTop");
                mainSprites[1] = GFX.SpriteBank.Create("WallBoosterMid");
                mainSprites[2] = GFX.SpriteBank.Create("WallBoosterBottom");
            }
        }

        public override void Added(Entity entity)
        {
            base.Added(entity);
            dyn = new DynamicData(type, entity);
            oldDepth = entity.Depth;
        }

        public override void Update()
        {
            base.Update();
            if (BetterIceWalls.Settings.Toggle != IceWallTypes.Disabled) {
                if (dyn.TryGet<List<Sprite>>("tiles", out List<Sprite> tiles) && dyn.TryGet<Facings>("Facing", out Facings facing))
                {
                    object o = dyn.Get("IceMode");
                    bool c = Entity.Get<ImageWrapper>() == null;
                    object p = dyn.Get("spriteOffset");
                    int offset;
                    if (p == null || !(p is int)) offset = 0;
                    else offset = (int)p - 1; //Bro why is the default value 1, i don't understand this is probably an error in ShroomHelper
                    for (int j = 0; j < tiles.Count; j++)
                    {
                        Sprite s = tiles[j];
                        if (o == null || (bool)o)
                        {

                            //Depth
                           
                            if (Entity.Get<StaticMover>().Platform != null) Entity.Depth = Entity.Get<StaticMover>().Platform.Depth - 1;
                            else Entity.Depth = Entity.CollideFirst<Solid>(Vector2.UnitX * (-4f * (float)facing))?.Depth - 1 ?? -10001;
                            //Changing Sprites
                            if ((int)BetterIceWalls.Settings.Toggle > 2) //Retexture
                            {
                                if (BetterIceWalls.Settings.Toggle == IceWallTypes.Optimized)
                                {
                                    s.Color = ColorCopy(s.Color, 0.1f);
                                    s.Scale = new Vector2(0.5f, 1f);
                                    if(facing == Facings.Left)
                                    {
                                        s.Position.X = 0f - offset;
                                    }
                                    else
                                    {
                                        s.Position.X = 8 - 4*s.Scale.X - offset;
                                    }
                                }
                                else
                                {
                                    s.Visible = false;
                                }
                                if (c)
                                {
                                    ImageWrapper i = new ImageWrapper(GFX.Game[$"BetterIceWalls/{facing}/iceSurface0{j % 2}"]);
                                    i.Color = s.Color * 0.333f;
                                    i.Scale = new Vector2(1f, 1f);
                                    if (facing == Facings.Left)
                                    {
                                        i.Position.X = -8f - offset;
                                    }
                                    else
                                    {
                                        i.Position.X = 8f - offset;
                                    }
                                    i.Position.Y = s.Position.Y;
                                    Entity.Add(i);
                                }
                            }
                            else if(BetterIceWalls.Settings.Toggle == IceWallTypes.Scaled)
                            {
                                s.Color = ColorCopy(s.Color, 0.4f);
                                s.Scale = new Vector2(0.5f, 1f);
                                if (facing == Facings.Left)
                                {
                                    s.Position.X = 0f - offset;
                                }
                                else
                                {
                                    s.Position.X = 8 - 4 * s.Scale.X - offset;
                                }
                            }
                            else //Shift Texture (replacing top and bottom with middle only because it looks better)
                            {
                                s.Scale = Vector2.One;
                                s.Color = ColorCopy(s.Color, 0.75f);
                                if(facing == Facings.Left)
                                {
                                    s.Position.X = -4;
                                }
                                else
                                {
                                    s.Position.X = 8;
                                }
                                if (s.GetAnimations() == mainSprites[0].GetAnimations() || s.GetAnimations() == mainSprites[2].GetAnimations())
                                {
                                    modifiedSprites[j] = s.GetAnimations();
                                    s.ClearAnimations();
                                    new DynData<Sprite>(s).Set<Dictionary<string, Sprite.Animation>>("animations", mainSprites[1].GetAnimations());

                                }
                            }
                        }
                        else
                        {
                            if (Entity.Depth != oldDepth) Entity.Depth = oldDepth;
                            
                            if ((int)BetterIceWalls.Settings.Toggle > 2) //Retexture
                            {
                                
                                if(Entity.Get<ImageWrapper>() != null) Entity.Remove(Entity.Components.GetAll<ImageWrapper>().ToArray());
                                if (BetterIceWalls.Settings.Toggle == IceWallTypes.Optimized)
                                {
                                    s.Color = ColorCopy(s.Color, 1f);
                                    if (facing == Facings.Left)
                                    {
                                        s.Position.X = 0f;
                                    }
                                    else
                                    {
                                        s.Position.X = 4f;
                                    }
                                    s.Scale = new Vector2(1f, 1f);
                                    
                                }
                                else
                                {
                                    s.Visible = true;
                                }
                            }
                            else if(BetterIceWalls.Settings.Toggle == IceWallTypes.Scaled)
                            {
                                if (facing == Facings.Left)
                                {
                                    s.Position.X = 0f;
                                }
                                else
                                {
                                    s.Position.X = 4f;
                                }
                                s.Scale = new Vector2(1f, 1f);
                                s.Color = ColorCopy(s.Color, 1f);
                            }
                            else //Shift Texture (replacing top and bottom with middle only because it looks better)
                            {
                                s.Position.X = 2 + (int)facing * 2;
                                s.Color = ColorCopy(s.Color, 1f);
                                if (modifiedSprites.ContainsKey(j))
                                {

                                    s.ClearAnimations();
                                    new DynData<Sprite>(s).Set<Dictionary<string, Sprite.Animation>>("animations", modifiedSprites[j]);
                                }
                            }
                        }
                        
                    }
                }
                
            }
            else
            {
                if (Entity.Depth != oldDepth) Entity.Depth = oldDepth;
                if (dyn.TryGet<List<Sprite>>("tiles", out List<Sprite> tiles) && dyn.TryGet<Facings>("Facing", out Facings facing)) {
                    for(int j = 0; j < tiles.Count; j++)
                    {
                        Sprite s = tiles[j];
                        s.Position.X = 2 + (int)facing * 2;
                        s.Scale = Vector2.One;
                        s.Color = ColorCopy(s.Color, 1f);
                        if (modifiedSprites.ContainsKey(j))
                        {
                            
                            s.ClearAnimations();
                            new DynData<Sprite>(s).Set<Dictionary<string, Sprite.Animation>>("animations", modifiedSprites[j]);
                        }
                    }
                }
            }
            
        }

        private Color ColorCopy(Color c, float alpha)
        {
            return new Color(c.R, c.G, c.B, alpha);
        }
    }
}
